use crate::{
    bundle::{BundleType, BundleValue, DynBundle},
    firrtl,
    intern::Interned,
    module::Module,
};
use clap::{Parser, Subcommand, ValueHint};
use std::{path::PathBuf, process::exit};

#[derive(Subcommand)]
enum CliCommand {
    /// Generate FIRRTL
    Firrtl {
        /// the directory to put the generated FIRRTL and associated files in
        #[arg(short, long, value_hint = ValueHint::DirPath)]
        output: PathBuf,
        /// the stem of the generated .fir file, e.g. to get foo.fir, pass --file-stem=foo
        #[arg(long)]
        file_stem: Option<String>,
    },
}

/// a simple CLI
///
/// Use like:
///
/// ```no_run
/// # use fayalite::{hdl_module, cli::Cli};
/// # #[hdl_module]
/// # fn my_module() {}
/// fn main() {
///     Cli::parse().run(my_module());
/// }
/// ```
///
/// You can also use it with a larger [`clap`]-based CLI like so:
///
/// ```no_run
/// # use fayalite::{hdl_module};
/// # #[hdl_module]
/// # fn my_module() {}
/// use clap::{Subcommand, Parser};
///
/// #[derive(Subcommand)]
/// pub enum Cmd {
///     #[command(flatten)]
///     Fayalite(fayalite::cli::Cli),
///     MySpecialCommand {
///         #[arg(long)]
///         foo: bool,
///     },
/// }
///
/// #[derive(Parser)]
/// pub struct Cli {
///     #[command(subcommand)]
///     cmd: Cmd, // or just use fayalite::cli::Cli directly
/// }
///
/// fn main() {
///     match Cli::parse().cmd {
///         Cmd::Fayalite(v) => v.run(my_module()),
///         Cmd::MySpecialCommand { foo } => println!("special: foo={foo}"),
///     }
/// }
/// ```
#[derive(Parser)]
// clear things that would be crate-specific
#[command(name = "Fayalite Simple CLI", about = None, long_about = None)]
pub struct Cli {
    #[command(subcommand)]
    subcommand: CliCommand,
}

impl clap::Subcommand for Cli {
    fn augment_subcommands(cmd: clap::Command) -> clap::Command {
        CliCommand::augment_subcommands(cmd)
    }

    fn augment_subcommands_for_update(cmd: clap::Command) -> clap::Command {
        CliCommand::augment_subcommands_for_update(cmd)
    }

    fn has_subcommand(name: &str) -> bool {
        CliCommand::has_subcommand(name)
    }
}

impl Cli {
    /// forwards to [`clap::Parser::parse()`] so you don't have to import [`clap::Parser`]
    pub fn parse() -> Self {
        clap::Parser::parse()
    }
    fn run_impl(self, top_module: Module<DynBundle>) -> ! {
        match self.subcommand {
            CliCommand::Firrtl { output, file_stem } => {
                let result = firrtl::export(
                    firrtl::FileBackend {
                        dir_path: output,
                        top_fir_file_stem: file_stem,
                    },
                    &top_module,
                );
                if let Err(e) = result {
                    eprintln!("Error: {:?}", eyre::Report::new(e));
                    exit(1);
                }
            }
        }
        exit(0)
    }
    pub fn run<T: BundleValue>(self, top_module: Interned<Module<T>>) -> !
    where
        T::Type: BundleType<Value = T>,
    {
        self.run_impl(top_module.canonical())
    }
}
